#!/usr/bin/perl -w
# 
# Copyright (c) BIO_MARK 2021
# Writer:         Dengdj <dengdj@biomarker.com.cn>
# Program Date:   2021
# Modifier:       Dengdj <dengdj@biomarker.com.cn>
# Last Modified:  2021
my $ver="1.3";

use strict;
use Getopt::Long;
use Data::Dumper;
use FindBin qw($Bin $Script);
use File::Basename qw(basename dirname);

#####################

my %opts;
GetOptions(\%opts,"i=s","k=s","b=s","o=s","m=s","s=s","l=s","h" );

#&help()if(defined $opts{h});
if(!defined($opts{i}) || !defined($opts{b}) || !defined($opts{k}) || !defined($opts{o}) || defined($opts{h}))
{
	print <<"	Usage End.";
	Description:
		
		Version: $ver
		v1.2   : add max offset for kmer map position vs ref kmer position, default <= 5 [-s]
		v1.3   : filter mult-bc reads type(like: bc1-bc2-bc2-bc3...).

	Usage:

		-i           reads file                              <infile>     must be given
		-k           kmer file                               <infile>     must be given
		-b           bc file                                 <infile>     must be given
		-o           result file                             <outfile>    must be given
		-m           min kmer number                         [int]        optional[3]  
		-s           max offset number                       [int]        optional[5]  
		-l           kmer length                             [int]        optional[8]  
		-h           Help document

	Usage End.

	exit;
}

###############Time
my $Time_Start;
$Time_Start = sub_format_datetime(localtime(time()));
&show_log("Start Time :[$Time_Start]");
################
$| = 1 ;
## get parameters
my $readsfile = $opts{i} ;
my $kmerfile = $opts{k} ;
my $bcfile = $opts{b} ;
my $outfile = $opts{o} ;
my $minkmer = defined $opts{m} ? $opts{m} : 3 ;
my $max_offset = defined $opts{s} ? $opts{s} : 5 ;
my $kmerlen = defined $opts{l} ? $opts{l} : 8 ;

my $umi_len = 12 ;
my %humi_pos = (
    'bc1' => 66,
    'bc2' => 44,
    'bc3' => 19,
);
my %href_pos = (
	'bc1'   => 1 ,
	'bc2'   => 23 ,
	'bc3'   => 48 ,
);

## get kmer
my %hkmer = ();
my %hkmer_pos = ();
&get_kmer($kmerfile, \%hkmer, \%hkmer_pos);

## get bc seq
my %hbc_seq = ();
my %hbc_len = ();
&get_bc_seq($bcfile, \%hbc_seq, \%hbc_len);

## process
&read_fq_and_check_barcode($readsfile, $outfile, $minkmer, $max_offset, \%hbc_seq, \%hbc_len, \%hkmer, \%hkmer_pos, $kmerlen, \%humi_pos, \%href_pos, $umi_len);

###############Time
my $Time_End;
$Time_End = sub_format_datetime(localtime(time()));
&show_log("End Time :[$Time_End]");

###############Subs
sub sub_format_datetime {#Time calculation subroutine
    my($sec, $min, $hour, $day, $mon, $year, $wday, $yday, $isdst) = @_;
	$wday = $yday = $isdst = 0;
    sprintf("%4d-%02d-%02d %02d:%02d:%02d", $year+1900, $mon+1, $day, $hour, $min, $sec);
}

sub ABSOLUTE_DIR
{ #$pavfile=&ABSOLUTE_DIR($pavfile);
	my $cur_dir=`pwd`;
	$cur_dir =~ s/\n$//;
	my ($in)=@_;
	my $return="";
	
	if(-f $in)
	{
		my $dir=dirname($in);
		my $file=basename($in);
		chdir $dir;$dir=`pwd`;
		$dir =~ s/\n$// ;
		$return="$dir/$file";
	}
	elsif(-d $in)
	{
		chdir $in;$return=`pwd`;
		$return =~ s/\n$// ;
	}
	else
	{
		warn "Warning just for file and dir [$in]\n";
		exit;
	}
	
	chdir $cur_dir;
	return $return;
}

# &show_log("txt")
sub show_log()
{
	my ($txt) = @_ ;
	my $time = time();
	my $Time = &sub_format_datetime(localtime($time));
	print "$Time:\t$txt\n" ;
	return ($time) ;
}

#&run_or_die($cmd);
sub run_or_die()
{
	my ($cmd) = @_ ;
	my $start_time = &show_log($cmd);
	my $flag = system($cmd) ;
	if ($flag != 0){
		my $end_time = &show_log("Error: command fail: $cmd");
		&show_log("Elaseped time: ".($end_time - $start_time)."s\n");
		exit(1);
	}
	my $end_time = &show_log("done.");
	&show_log("Elaseped time: ".($end_time - $start_time)."s\n");
	return ;
}

## qsub
sub qsub()
{
	my ($shfile, $maxproc) = @_ ;
	my $dir = dirname($shfile) ;
	my $cmd = "cd $dir && sh /share/nas2/genome/bmksoft/tool/qsub_sge_plus/v1.0/qsub_sge.plus.sh --maxproc $maxproc --reqsub --independent $shfile" ;
	&run_or_die($cmd);

	return ;
}

## qsub_mult($shfile, $max_proc, $job_num)
sub qsub_mult()
{
	my ($shfile, $max_proc, $job_num) = @_ ;
	if ($job_num > 500){
		my @shfiles = &cut_shfile($shfile);
		for my $file (@shfiles){
			&qsub($file, $max_proc);
		}
	}
	else{
		&qsub($shfile, $max_proc) ;
	}
}

#my @shfiles = &cut_shfile($shfile);
sub cut_shfile()
{
	my ($file) = @_ ;
	my @files = ();
	my $num = 0 ;
	open (IN, $file) || die "Can't open $file, $!\n" ;
	(my $outprefix = $file) =~ s/.sh$// ;
	while (<IN>){
		chomp ;
		if ($num % 500 == 0){
			close(OUT);
			my $outfile = "$outprefix.sub_".(int($num/500)+1).".sh" ;
			push @files, $outfile ;
			open (OUT, ">$outfile") || die "Can't creat $outfile, $!\n" ;
		}
		print OUT $_, "\n" ;
		$num ++ ;
	}
	close(IN);
	close(OUT);

	return @files ;
}

#&get_kmer($kmerfile, \%hkmer, \%hpos);
sub get_kmer()
{
	my ($kmerfile, $ahkmer, $ahpos) = @_ ;

	open (IN, $kmerfile) || die "$kmerfile, $!\n" ;
	while(<IN>){
		chomp ;
		next if (m/^\s*$|^\#/);
		my ($kmer, $num, @ids) = split ;
		for (my $i=0; $i<@ids; $i+=2){
			push @{$ahkmer->{$kmer}}, $ids[$i] ;
			push @{$ahpos->{$kmer}}, $ids[$i+1] ;
		}
	}
	close(IN);

	return ;
}

#&get_bc_seq($bcfile, \%hbc_seq, \%hbc_len);
sub get_bc_seq()
{
    my ($bcfile, $ahbc_seq, $ahbc_len) = @_ ;

    open (IN, $bcfile) || die "$bcfile, $!\n" ;
    $/ = ">" ;
    while(<IN>){
        chomp ;
        next if (m/^\s*$/);
        my ($id, $seq) = split(/\n/, $_, 2) ;
        $id = (split /\s+/, $id)[0] ;
        $seq =~ s/\n//g ;
        my $len = length($seq) ;
        $ahbc_seq->{$id} = $seq ;
        $ahbc_len->{$id} = $len ;
    }
    $/ = "\n" ;
    close(IN);

    return ;
}

#&read_fq_and_check_barcode($fqfile, $outfile, $minkmer, $max_offset, \%hbc_seq, \%hbc_len, \%hkmer, \%hkmer_pos, $kmerlen, \%humi_pos, \%href_pos, $umi_len);
sub read_fq_and_check_barcode()
{
	my ($fqfile, $outfile, $minkmer, $max_offset, $ahbc_seq, $ahbc_len, $ahkmer, $ahkmer_pos, $kmerlen, $ahumi_pos, $ahref_pos, $umi_len) = @_ ;

    my $str = $fqfile ;
    if ($fqfile =~ /.gz$/){
        $str = "gunzip -c $fqfile |" ;
    }
	open (IN, $str) || die "$fqfile, $!\n" ;
	open (OUT, ">$outfile") || die "$outfile, $!\n" ;
	while(<IN>){
		my $id = $_ ;
		my $seq = <IN> ;
		my $id2 = <IN> ;
		my $qual = <IN> ;
		chomp($id, $seq, $id2, $qual);
		$id = (split /\s+/, $id)[0] ;
		$id =~ s/^\@// ;
        $id =~ s/\/1$// ;
		my $len = length($seq) ;
        my $offlen = $len > 75 ? 75 : $len ;
		my %htypes = ();
        my %hkmer_ref = ();
		for (my $i=0; $i<$offlen-$kmerlen+1; $i++){
			my $kmer = substr($seq, $i, $kmerlen);
			if (defined $ahkmer->{$kmer}){
                for (my $j=0; $j<@{$ahkmer->{$kmer}}; $j++){
                    my $ref_pos = $ahkmer_pos->{$kmer}[$j] ;
                    my $map_pos = $i + 1;
                    my $offset = abs($ref_pos - $map_pos) ;
                    if ($offset > $max_offset){
                        next ;
                    }
                    if ($ahkmer->{$kmer}[$j] =~ m/(bc\d)\_(\d+)/){
                        $htypes{$1}{$ahkmer->{$kmer}[$j]}++ ;
                    }
                    push @{$hkmer_ref{$ahkmer->{$kmer}[$j]}}, [$ahkmer_pos->{$kmer}[$j], $map_pos] ;
                }
            }
		}
        my %hbc_type = ();
        my $umi_seq = "" ;
        my @types = ();
        my @cigars = ();
        my $bc_flag = 0 ;
        for my $type (sort keys %htypes){
            for my $bc (sort {$htypes{$type}{$b} <=> $htypes{$type}{$a}} keys %{$htypes{$type}}){
                my $bc_seq = $ahbc_seq->{$bc} ;
                my $bc_len = $ahbc_len->{$bc} ;
                my @poses = @{$hkmer_ref{$bc}} ;
                my @bc_reg = ([$poses[0][0], $poses[0][0]]);
                my @read_reg = ([$poses[0][1], $poses[0][1]]);
                for (my $i=1; $i<@poses; $i++){
                    my ($ref_pos, $read_pos) = @{$poses[$i]} ;
                    if ($ref_pos - $bc_reg[-1][1] == 1 && $read_pos - $read_reg[-1][1] == 1){
                        $bc_reg[-1][1] = $ref_pos ;
                        $read_reg[-1][1] = $read_pos ;
                    }
                    else{
                        push @bc_reg, [$ref_pos, $ref_pos] ;
                        push @read_reg, [$read_pos, $read_pos] ;
                    }
                }
                #print $bc , "\n" ;
                #print Dumper @read_reg, "\n" ;
                #print Dumper @bc_reg, "\n" ;
                my ($flag, $match_num, $mismatch_num, $cigar, $bc_end_pos, $read_end_pos) = &judge_match(\@bc_reg, \@read_reg, $bc_seq, $bc_len, $seq, $kmerlen, $ahref_pos->{$type});
                if ($flag == 1){
                    my $umi_pos = $ahumi_pos->{$type} + ($read_end_pos - $bc_end_pos) + $ahref_pos->{$type} ;
                    my $umi_seq = substr($seq, $umi_pos-1, $umi_len);
                    $hbc_type{$type} = [$bc, $cigar, $bc_end_pos, $read_end_pos, $umi_pos, $umi_seq] ;
                    push @types, $bc ;
                    push @cigars, $cigar ;
                    $bc_flag = 1 ;
                    last ;
                }
            }
        }
        if ($bc_flag == 1){
            print OUT "$id\t", join("-", @types),"\t", join(":",@cigars) ;
            for my $type (sort keys %hbc_type){
                my ($bc, $cigar, $bc_end_pos, $read_end_pos, $umi_pos, $umi_seq) = @{$hbc_type{$type}} ;
                print OUT "\t$bc_end_pos:$read_end_pos:$umi_pos\t$umi_seq" ;
            }
            print OUT "\n" ;
        }
        else{
            print OUT "$id\tNull\n" ;
        }
	}
	close(IN);
	close(OUT);

	return ;
}

#my ($flag, $match_num, $mismatch_num, $cigar, $bc_end_pos, $read_end_pos) = &judge_match(\@bc_reg, \@read_reg, $bc_seq, $bc_len, $seq, $kmerlen, $ahref_pos->{$type});
sub judge_match()
{
    my ($abc_reg, $aread_reg, $bc_seq, $bc_len, $seq, $kmerlen, $ref_start) = @_ ;

    for (my $i=0; $i<@{$abc_reg}; $i++){
        my ($bc_start, $bc_end) = @{$abc_reg->[$i]} ;
        my ($read_start, $read_end) = @{$aread_reg->[$i]} ;
        $bc_end += $kmerlen-1 ;
        $read_end += $kmerlen-1 ;
        my $match = $bc_end - $bc_start + 1 ;
        if ($match >= $bc_len-1){
            my $cigar = $bc_len."=" ;
            if ($bc_start > $ref_start){
                $cigar = "1X".($bc_len-1)."=" ;
            }
            elsif ($bc_end < $bc_len+$ref_start-1){
                $cigar = ($bc_len-1)."=1X" ;
            }
            return (1, $match, $bc_len - $match, $cigar, $bc_end, $read_end) ;
        }
        if ($bc_start > $ref_start && $bc_end < $bc_len+$ref_start-1){
            next ;
        }
        my ($mismatch, $cigar, $bc_end_pos, $read_end_pos) = &reg_match($bc_start, $bc_end, $read_start, $read_end, $bc_seq, $bc_len, $seq, $ref_start);
        if ($mismatch == 1){
            return (1, $bc_len-1, 1, $cigar, $bc_end_pos, $read_end_pos);
        }
    }

    return(0);
}

#my ($mismatch, $cigar, $bc_end_pos, $read_end_pos) = &reg_match($bc_start, $bc_end, $read_start, $read_end, $bc_seq, $bc_len, $seq, $ref_start);
sub reg_match()
{
    my ($bc_start, $bc_end, $read_start, $read_end, $bc_seq, $bc_len, $seq, $ref_start) = @_ ;

    my $mismatch = -1 ;
    my $cigar = "" ;
    my $bc_end_pos = $bc_end ;
    my $read_end_pos = $read_end ;
    if ($bc_start > $ref_start){
        my $bc_pos_start = $bc_start - $ref_start ;
        my $bc_cut_sq1 = substr($bc_seq, 0, $bc_pos_start-1);
        my $bc_cut_sq2 = substr($bc_seq, 0, $bc_pos_start-2);
        my $read_cut_sq1 = substr($seq, $read_start-$bc_pos_start-1, $bc_pos_start-1);
        my $read_cut_sq2 = substr($seq, $read_start-$bc_pos_start, $bc_pos_start-2);
        my $read_cut_sq3 = substr($seq, $read_start-$bc_pos_start+1, $bc_pos_start-2);
        if ($bc_cut_sq2 eq $read_cut_sq2){
            $mismatch = 1 ;
            $cigar = ($bc_pos_start-2)."=1X".($bc_end-$bc_start+1)."=" ;
        }
        elsif ($bc_cut_sq2 eq $read_cut_sq3){
            $mismatch = 1 ;
            $cigar = ($bc_pos_start-2)."=1I".($bc_end-$bc_start+1)."=" ;
        }
        elsif ($bc_cut_sq1 eq $read_cut_sq1){
            $mismatch = 1 ;
            $cigar = ($bc_pos_start-1)."=1D".($bc_end-$bc_start+1)."=" ;
        }
        else{
            return(-1);
        }
    }
    elsif ($bc_end < $ref_start+$bc_len-1){
        #print "$bc_seq\n" ;
        #print "$seq\n" ;
        #print "$bc_end\t$read_end\n" ;
        my $bc_pos_end = $bc_end - $ref_start + 1 ;
        my $bc_cut_sq1 = substr($bc_seq, $bc_pos_end, $bc_len-$bc_pos_end);
        my $bc_cut_sq2 = substr($bc_seq, $bc_pos_end+1, $bc_len-$bc_pos_end-1);
        my $read_cut_sq1 = substr($seq, $read_end+1, $bc_len-$bc_pos_end-1);
        my $read_cut_sq2 = substr($seq, $read_end, $bc_len-$bc_pos_end-1);
        my $read_cut_sq3 = substr($seq, $read_end+1, $bc_len-$bc_pos_end);
        #print "$bc_cut_sq1\n" ;
        #print "$bc_cut_sq2\n" ;
        #print "$read_cut_sq1\n" ;
        #print "$read_cut_sq2\n" ;
        #print "$read_cut_sq3\n" ;
        if ($bc_cut_sq2 eq $read_cut_sq1){
            $mismatch = 1 ;
            $cigar = "$bc_pos_end=1X".($bc_len-$bc_pos_end-1)."=" ;
            $bc_end_pos = $ref_start + $bc_len - 1 ;
            $read_end_pos = $read_end + $bc_len - $bc_pos_end ;
        }
        elsif ($bc_cut_sq1 eq $read_cut_sq3){
            $mismatch = 1 ;
            $cigar = "$bc_end=1D".($bc_len-$bc_end)."=" ;
            $bc_end_pos = $ref_start + $bc_len - 1 ;
            $read_end_pos = $read_end + $bc_len - $bc_pos_end + 1 ;
        }
        elsif ($bc_cut_sq2 eq $read_cut_sq2){
            $mismatch = 1 ;
            $cigar = "$bc_end=1I".($bc_len-$bc_end-1)."=" ;
            $bc_end_pos = $ref_start + $bc_len - 1 ;
            $read_end_pos = $read_end + $bc_len - $bc_pos_end - 1 ;
        }
        else{
            return(-1);
        }
    }

    return($mismatch, $cigar, $bc_end_pos, $read_end_pos);
}


