library("optparse")

option_list <- list(
	make_option(c("-i", "--indir"), help="indir"),
	make_option(c("-o", "--outdir"), help="outdir", default="."),
	make_option(c("--he_png"), help="he png file"),
        make_option(c("--prefix"), help="prefix"),
	make_option(c("--res"), help="resolution, default %default",type="double",default=0.5),
	make_option(c("--MinCell"), help="MinCell, default %default",type="double",default=5),
	make_option(c("--MinFeatures"), help="MinFeatures, default %default",type="double",default=100),
	make_option(c("--point_size"), help="point_size, default %default",type="double",default=1.7)
)
opt <- parse_args(OptionParser(option_list=option_list))
#######################################
# 首先将需要用到的包load进工作环境
suppressMessages({
  library(Seurat)
  library(dplyr)
  library(tibble)
  library(ggplot2)
  #library(tiff)
  library(broom)
  #library(grid)
  #library(gridExtra)
  #library(magick)
  library(purrr)
  library(cowplot)
  #library(figpatch)
  #library(pdftools)
  library(cluster)
  library(ggpubr)
  library(plotly)
  library(htmlwidgets)
})

# 记录运行时间
StartTime=proc.time()

##绘图函数
#输出ggplot对象
SavePlot <- function(od, filename, data, width = 8, height = 8){
  file.png <- paste(filename, "png", sep = ".")
  file.pdf <- paste(filename, "pdf", sep = ".")
  ggsave(filename = file.path(od, file.png), plot = data, width = width, height = height)
  ggsave(filename = file.path(od, file.pdf), plot = data, width = width, height = height)
}



# 参数初始化
#选择不同的resolution值可以获得不同的cluster数目，值越大cluster数目越多，默认值是0.5
res=opt$res
FilePath=opt$indir
he_file=opt$he_png
#过滤参数：细胞数5，基因数100
MinCell=opt$MinCell
MinFeatures=opt$MinFeatures

#输出目录
outdir=opt$outdir
if(!dir.exists(outdir)){dir.create(path=outdir)}


# 数据导入，给定输入文件路径，路径下需包括matrix.mtx,genes.tsv(或 features.tsv),和barcodes.tsv文件
expr=Read10X(FilePath,gene.column = 1)

#创建Seurat对象，数据集中测到的少于200个基因的位点（min.features = 200）和少于5个位点覆盖的基因（min.cells = 3）被过滤掉
object=CreateSeuratObject(counts = expr,assay = "Spatial",min.cells=MinCell,min.features=MinFeatures)
#标准化，鉴定出表达量高变的基因
#object <- SCTransform(object,assay = "Spatial",variable.features.n = 2000) 
object <- NormalizeData(object, normalization.method = "LogNormalize", scale.factor = 10000)
object <- FindVariableFeatures(object, selection.method = "vst", nfeatures = 2000)
object <- ScaleData(object)

object <- RunPCA(object=object,pc.genes = VariableFeatures(object))
object <- FindNeighbors(object=object, reduction = "pca",dims = 1:30,verbose = F)
object <- FindClusters(object=object, verbose = F,resolution = res)
object <- RunUMAP(object=object, reduction = "pca",dims = 1:30)

#整理聚类结果
Cluster=object@meta.data %>%
  select(.,seurat_clusters) %>%
  rownames_to_column(.,var = "Barcode") %>%
  arrange(.,seurat_clusters)

#输出聚类结果
write.csv(Cluster,file.path(outdir,"cluster.csv"),quote = F,row.names = F)

# 计录聚类时间
EndTime=proc.time()
RunTime <- EndTime - StartTime
print(paste0('ProcessingTime: ',RunTime[3][[1]],'seconds'))


##找到各个类别的marker gene
#object.markers <- FindAllMarkers(object, only.pos = TRUE, min.pct = 0.25)

##提取每个类别的marker gene，n=10表示提取每个类别中top10的marker gene
#top10 <- object.markers %>% group_by(cluster) %>% top_n(n = 10, wt = avg_log2FC)

##将每个类别中的marker gene输出
#write.csv(top10,file.path(outdir,"marker_gene.csv"),quote = F, row.names = F)

#Cluster=object@meta.data %>%
#  select(.,c(seurat_clusters)) %>%
#  rownames_to_column(.,var = "Barcode") %>%
#  arrange(.,seurat_clusters)

col = c("#F56867","#FEB915","#C798EE","#59BE86","#7495D3","#D1D1D1","#6D1A9C","#15821E","#3A84E6","#70e014","#787878","#DB4C6C","#0430e0","#554236","#AF5F3C","#ff7700","#e00417","#DAB370","#fcfc05","#268785","#ed1299", "#09f9f5", "#246b93", "#cc8e12", "#d561dd", "#c93f00", "#ddd53e","#4aef7b", "#e86502", "#9ed84e", "#39ba30", "#6ad157", "#8249aa", "#99db27", "#e07233", "#ff523f","#ce2523", "#f7aa5d", "#cebb10", "#03827f", "#931635", "#373bbf", "#a1ce4c", "#ef3bb6", "#d66551","#1a918f", "#ff66fc", "#2927c4", "#7149af" ,"#57e559" ,"#8e3af4" ,"#f9a270" ,"#22547f", "#db5e92","#edd05e", "#6f25e8", "#0dbc21", "#280f7a", "#6373ed", "#5b910f" ,"#7b34c1" ,"#0cf29a" ,"#d80fc1", "#51f59b","#FF0000", "#FF4D00", "#FF9900", "#FFE500", "#CCFF00", "#80FF00", "#33FF00", "#00FF19", "#00FF66", "#00FFB2", "#00FFFF", "#00B3FF", "#0066FF", "#001AFF", "#3300FF", "#7F00FF", "#CC00FF", "#FF00E6", "#FF0099", "#FF004D")
UMAP_df <- object[['umap']]@cell.embeddings %>%
  as.data.frame() %>%
  rownames_to_column(var = 'Barcode') %>%
  left_join(Cluster, by = 'Barcode') %>%
  dplyr::rename(cluster = seurat_clusters) %>%
  select(Barcode, cluster, everything())
UMAP_df$cluster <- as.factor(UMAP_df$cluster)

umap_pic <- ggplot(UMAP_df, aes(x = UMAP_1, y = UMAP_2))+
      geom_point(aes(color = cluster),size=opt$point_size, shape=16)+
      scale_color_manual(values = col)+
      labs(colour = "cluster")+
      theme_bw(base_size = 15)+
      guides(colour = guide_legend(override.aes = list(size=3.5)))
#pdf(paste(outdir,'umap.pdf',sep="/"), width = 12/3.8*1.8, height = 5)
if(is.null(opt$prefix)){
    ggsave(filename = paste(outdir,"umap.pdf",sep="/"), plot = umap_pic, width = 10, height = 8)
    ggsave(filename = paste(outdir,"umap.png",sep="/"), plot = umap_pic, width = 10, height = 8)
    htmlwidgets::saveWidget(as_widget(ggplotly(umap_pic)), file = paste(outdir,'umap.html',sep="/"))
}else{
    ggsave(filename = paste(outdir,"/",opt$prefix,"_umap.pdf",sep=""), plot = umap_pic, width = 10, height = 8)
    ggsave(filename = paste(outdir,"/",opt$prefix,"_umap.png",sep=""), plot = umap_pic, width = 10, height = 8)
    htmlwidgets::saveWidget(as_widget(ggplotly(umap_pic)), file = paste(outdir,"/",opt$prefix,"_umap.html",sep=""))
}

#htmlwidgets::saveWidget(as_widget(ggplotly(umap_pic)), file = paste(outdir,'umap.html',sep="/"))

he_fig=png::readPNG(he_file)
width=ncol(he_fig)
hetght=nrow(he_fig)
cal_zoom_rate = function(width, height){
	std_width = 1000
    	std_height = std_width / (46 * 31) * (46 * 36 * sqrt(3) / 2.0)
    	if(std_width / std_height > width / height){
        	scale = width / std_width
	}
    	else{
        	scale = height / std_height
	}
    	return(scale)
}

zoom_scale = cal_zoom_rate(ncol(he_fig), nrow(he_fig))

bc_pos_file = gzfile(paste(FilePath, "barcodes_pos.tsv.gz", sep  = "/"),'rt')
bc_pos = read.table(bc_pos_file, header = FALSE, sep = '\t', quote = '')
names(bc_pos)=c("Barcode","x","y")
bc_pos=merge(bc_pos,Cluster,by="Barcode",all=FALSE)
bc_pos <- bc_pos %>%
  dplyr::rename(cluster = seurat_clusters)

bc_pos$x=bc_pos$x * zoom_scale
bc_pos$y=bc_pos$y * zoom_scale

bc_pos %>%
    group_by(cluster) %>% 
    do(model = pam(.[c('x', 'y')], 1, stand = T)) %>% 
    ungroup() %>% group_by(cluster) %>% 
    do(map_df(.$model, broom::tidy)) %>% ### 整理模型数据
    ungroup() %>% select(cluster, x,y) %>% 
    data.frame() %>% 
    dplyr::rename(x.center = x, y.center = y, cluster = cluster) -> label.data

clstr_plot<-ggplot(data=bc_pos,aes(x ,dim(he_fig)[1]-y)) + 
    background_image(he_fig)+
    geom_point(aes(colour = cluster),size = opt$point_size, shape=16) + 
    scale_color_manual(values = col)+
    theme_bw() +
    theme(plot.title = element_text(face = 2,size = 50,hjust = 0.5)) +
    theme(axis.ticks = element_blank(), axis.text.y = element_blank(),panel.border = element_blank(), axis.text.x = element_blank()) +
    xlab('')+ylab('')+
    coord_cartesian(xlim = c(0, dim(he_fig)[2]), ylim = c(0, dim(he_fig)[1]), 
                        expand = FALSE)+
    guides(colour = guide_legend(override.aes = list(size=3.5)))+
    theme(legend.position="right")

if(is.null(opt$prefix)){
    ggsave(filename = paste(outdir,"cluster.pdf",sep="/"), plot = clstr_plot, width = 9, height = 8)
    ggsave(filename = paste(outdir,"cluster.png",sep="/"), plot = clstr_plot, width = 9, height = 8)
    htmlwidgets::saveWidget(as_widget(ggplotly(clstr_plot)), file = paste(outdir,'cluster.html',sep="/"))
}else{
    ggsave(filename = paste(outdir,"/",opt$prefix,"_cluster.pdf",sep=""), plot = clstr_plot, width = 9, height = 8)
    ggsave(filename = paste(outdir,"/",opt$prefix,"_cluster.png",sep=""), plot = clstr_plot, width = 9, height = 8)
    htmlwidgets::saveWidget(as_widget(ggplotly(clstr_plot)), file = paste(outdir,"/",opt$prefix,"_cluster.html",sep=""))
}
#htmlwidgets::saveWidget(as_widget(ggplotly(clstr_plot)), file = paste(outdir,'cluster.html',sep="/"))

p<-plot_grid(umap_pic,clstr_plot)
if(is.null(opt$prefix)){
    ggsave(filename = paste(outdir,"umap_clstr.pdf",sep="/"), plot = p, width = 18, height = 8)
    ggsave(filename = paste(outdir,"umap_clstr.png",sep="/"), plot = p, width = 18, height = 8)
}else{
    ggsave(p, filename = paste(outdir,"/",opt$prefix,"_umap_clstr.png",sep=""),height=8,width=18)
    ggsave(p, filename = paste(outdir,"/",opt$prefix,"_umap_clstr.pdf",sep=""),height=8,width=18)
}

clstr_plot<-ggplot(data=bc_pos,aes(x ,dim(he_fig)[1]-y)) + 
    #background_image(he_fig)+
    geom_point(aes(colour = cluster),size = opt$point_size, shape=16) + 
    scale_color_manual(values = col)+
    theme_bw() +
    theme(plot.title = element_text(face = 2,size = 50,hjust = 0.5)) +
    theme(axis.ticks = element_blank(), axis.text.y = element_blank(),panel.border = element_blank(), axis.text.x = element_blank()) +
    xlab('')+ylab('')+
    coord_cartesian(xlim = c(0, dim(he_fig)[2]), ylim = c(0, dim(he_fig)[1]), 
                        expand = FALSE)+
    guides(colour = guide_legend(override.aes = list(size=3.5)))+
    theme(legend.position="right")

if(is.null(opt$prefix)){
    ggsave(filename = paste(outdir,"cluster_wt.pdf",sep="/"), plot = clstr_plot, width = 9, height = 8)
    ggsave(filename = paste(outdir,"cluster_wt.png",sep="/"), plot = clstr_plot, width = 9, height = 8)
    htmlwidgets::saveWidget(as_widget(ggplotly(clstr_plot)), file = paste(outdir,'cluster.html',sep="/"))
}else{
    ggsave(filename = paste(outdir,"/",opt$prefix,"_cluster_wt.pdf",sep=""), plot = clstr_plot, width = 9, height = 8)
    ggsave(filename = paste(outdir,"/",opt$prefix,"_cluster_wt.png",sep=""), plot = clstr_plot, width = 9, height = 8)
    htmlwidgets::saveWidget(as_widget(ggplotly(clstr_plot)), file = paste(outdir,"/",opt$prefix,"_cluster.html",sep=""))
}
#htmlwidgets::saveWidget(as_widget(ggplotly(clstr_plot)), file = paste(outdir,'cluster.html',sep="/"))

p<-plot_grid(umap_pic,clstr_plot)
if(is.null(opt$prefix)){
    ggsave(filename = paste(outdir,"umap_clstr_wt.pdf",sep="/"), plot = p, width = 18, height = 8)
    ggsave(filename = paste(outdir,"umap_clstr_wt.png",sep="/"), plot = p, width = 18, height = 8)
}else{
    ggsave(p, filename = paste(outdir,"/",opt$prefix,"_umap_clstr_wt.png",sep=""),height=8,width=18)
    ggsave(p, filename = paste(outdir,"/",opt$prefix,"_umap_clstr_wt.pdf",sep=""),height=8,width=18)
}

