#!/usr/bin/perl -w
# 
# Copyright (c) BIO_MARK 2023
# Writer:         Dengdj <dengdj@biomarker.com.cn>
# Program Date:   2023
# Modifier:       Dengdj <dengdj@biomarker.com.cn>
# Last Modified:  2023
my $ver="1.0";

use strict;
use Getopt::Long;
use Data::Dumper;
use FindBin qw($Bin $Script);
use File::Basename qw(basename dirname);

#####################

my %opts;
GetOptions(\%opts,"u=s","b=s","c=s","o=s", "n=s","h" );

#&help()if(defined $opts{h});
if(!defined($opts{u}) || !defined($opts{b}) || !defined($opts{c}) || !defined($opts{n}) || !defined($opts{o}) || defined($opts{h}))
{
	print <<"	Usage End.";
	Description:
		
		Version: $ver

	Usage:

		-u           umi_gene.tsv.gz                <infile>     must be given
		-b           barcode.tsv                    <infile>     must be given
		-c           barcodes_cluster.tsv.gz        <infile>     must be given
		-n           raw reads number               <int>        must be given
		-o           result file                    <outfile>    must be given
		-h           Help document

	Usage End.

	exit;
}

###############Time
my $Time_Start;
$Time_Start = sub_format_datetime(localtime(time()));
&show_log("Start Time :[$Time_Start]");
################
$| = 1 ;
## get parameters
my $umi_gene_file = $opts{u} ;
my $bc_file = $opts{b} ;
my $clust_file = $opts{c} ;
my $outfile = $opts{o} ;
my $total_reads = $opts{n} ;

## barcode.tsv
&show_log("reading $bc_file");
my %hbc2index = ();
my %hindex2bc = ();
&get_barcode_index($bc_file,\%hbc2index, \%hindex2bc);

## get supspot relation
&show_log("reading $clust_file");
my %hsup = ();
&get_supspot_relation($clust_file, \%hindex2bc, \%hsup);

## get bc-umi-gene-reads
&show_log("reading $umi_gene_file");
my %hbc_gene = ();
my $sum_reads = &get_umi_gene($umi_gene_file, \%hbc2index, \%hbc_gene);

## stat and output
&show_log("stat and output");
&stat_and_output($total_reads, $sum_reads, \%hsup, \%hbc_gene, $outfile);

###############Time
my $Time_End;
$Time_End = sub_format_datetime(localtime(time()));
&show_log("End Time :[$Time_End]");

###############Subs
sub sub_format_datetime {#Time calculation subroutine
    my($sec, $min, $hour, $day, $mon, $year, $wday, $yday, $isdst) = @_;
	$wday = $yday = $isdst = 0;
    sprintf("%4d-%02d-%02d %02d:%02d:%02d", $year+1900, $mon+1, $day, $hour, $min, $sec);
}

sub ABSOLUTE_DIR
{ #$pavfile=&ABSOLUTE_DIR($pavfile);
	my $cur_dir=`pwd`;
	$cur_dir =~ s/\n$//;
	my ($in)=@_;
	my $return="";
	
	if(-f $in)
	{
		my $dir=dirname($in);
		my $file=basename($in);
		chdir $dir;$dir=`pwd`;
		$dir =~ s/\n$// ;
		$return="$dir/$file";
	}
	elsif(-d $in)
	{
		chdir $in;$return=`pwd`;
		$return =~ s/\n$// ;
	}
	else
	{
		warn "Warning just for file and dir [$in]\n";
		exit;
	}
	
	chdir $cur_dir;
	return $return;
}

# &show_log("txt")
sub show_log()
{
	my ($txt) = @_ ;
	my $time = time();
	my $Time = &sub_format_datetime(localtime($time));
	print "$Time:\t$txt\n" ;
	return ($time) ;
}

#&run_or_die($cmd);
sub run_or_die()
{
	my ($cmd) = @_ ;
	my $start_time = &show_log($cmd);
	my $flag = system($cmd) ;
	if ($flag != 0){
		my $end_time = &show_log("Error: command fail: $cmd");
		&show_log("Elaseped time: ".($end_time - $start_time)."s\n");
		exit(1);
	}
	my $end_time = &show_log("done.");
	&show_log("Elaseped time: ".($end_time - $start_time)."s\n");
	return ;
}

## qsub
sub qsub()
{
	my ($shfile, $maxproc) = @_ ;
	my $dir = dirname($shfile) ;
	my $cmd = "cd $dir && sh /share/nas2/genome/bmksoft/tool/qsub_sge_plus/v1.0/qsub_sge.plus.sh --maxproc $maxproc --reqsub --independent $shfile" ;
	&run_or_die($cmd);

	return ;
}

## qsub_mult($shfile, $max_proc, $job_num)
sub qsub_mult()
{
	my ($shfile, $max_proc, $job_num) = @_ ;
	if ($job_num > 500){
		my @shfiles = &cut_shfile($shfile);
		for my $file (@shfiles){
			&qsub($file, $max_proc);
		}
	}
	else{
		&qsub($shfile, $max_proc) ;
	}
}

#my @shfiles = &cut_shfile($shfile);
sub cut_shfile()
{
	my ($file) = @_ ;
	my @files = ();
	my $num = 0 ;
	open (IN, $file) || die "Can't open $file, $!\n" ;
	(my $outprefix = $file) =~ s/.sh$// ;
	while (<IN>){
		chomp ;
		if ($num % 500 == 0){
			close(OUT);
			my $outfile = "$outprefix.sub_".(int($num/500)+1).".sh" ;
			push @files, $outfile ;
			open (OUT, ">$outfile") || die "Can't creat $outfile, $!\n" ;
		}
		print OUT $_, "\n" ;
		$num ++ ;
	}
	close(IN);
	close(OUT);

	return @files ;
}

sub sub_normal_dis_ran(){#$average,$standard_deviation
        my ($aver,$stand_dev) = @_ ;
        my $ran1 = rand() ;
        my $ran2 = rand() ;
        my $y = ((-2*log($ran1))**(0.5))*sin(2*3.1415926535*$ran2) ;
        return ($aver + $stand_dev*$y) ;
}

#&get_barcode_index($bc_file,\%hbc2index, \%hindex2bc);
sub get_barcode_index()
{
    my ($bcfile, $ahbc2index, $ahindex2bc) = @_ ;

    if ($bcfile =~ m/.gz$/){
        open (IN, "gunzip -c $bcfile|") || die "$bcfile, $!\n" ;
    }
    else{
        open (IN, $bcfile) || die "$bcfile, $!\n" ;
    }
    my $index = 0 ;
    while(<IN>){
        chomp ;
        my ($bc) = split ;
        $index++ ;
        $ahbc2index->{$bc} = $index ;
        $ahindex2bc->{$index} = $bc ;
    }
    close(IN);

    return ;
}

#&get_supspot_relation($clust_file, \%hindex2bc, \%hsup);
sub get_supspot_relation()
{
    my ($clust_file, $ahindex2bc, $ahsup) = @_ ;

    if ($clust_file =~ m/.gz$/){
        open (IN, "gunzip -c $clust_file |") || die "$clust_file , $!\n" ;
    }
    else{
        open (IN, $clust_file) || die "$clust_file, $!\n" ;
    }
    while(<IN>){
        chomp ;
        my ($line_num, $supspot) = split ;
        $supspot -= 1 ;
        my $bc = $ahindex2bc->{$line_num} ;
        $ahsup->{$supspot}{$bc} = 1 ;
    }
    close(IN);

    return ;
}

#my $sum_reads = &get_umi_gene($umi_gene_file, \%hbc2index, \%hbc_gene);
sub get_umi_gene()
{
    my ($umi_gene_file, $ahbc2index, $ahbc_gene) = @_ ;

    if ($umi_gene_file =~ m/.gz$/){
        open (IN, "gunzip -c $umi_gene_file |") || die "$umi_gene_file, $!\n" ;
    }
    else{
        open (IN, $umi_gene_file) || die "$umi_gene_file, $!\n" ;
    }
    my $sum = 0 ;
    while(<IN>){
        chomp ;
        my ($bc, $umi, $gene, $reads) = split ;
        next if (!defined $ahbc2index->{$bc});
        $ahbc_gene->{$bc}{$umi} = [$gene, $reads] ;
        $sum += $reads ;
    }
    close(IN);

    return($sum) ;
}

#&stat_and_output($total_reads, $sum_reads, \%hsup, \%hbc_gene, $outfile);
sub stat_and_output()
{
    my ($total_reads, $sum_reads, $ahsup, $ahbc_gene, $outfile) = @_ ;

    my %humi_stat = ();
    my %hgene_stat = ();
    my %hreads_stat = ();
    my $process_reads = 0 ;
    my $perc = 0 ;
    for my $sup (keys %{$ahsup}){
        for my $bc (keys %{$ahsup->{$sup}}){
            next if (!defined $ahbc_gene->{$bc}) ;
            for my $umi (keys %{$ahbc_gene->{$bc}}){
                my ($gene, $reads) = @{$ahbc_gene->{$bc}{$umi}} ;
                for (my $i=0; $i<$reads; $i++){
                    my $rand = rand(100);
                    for (my $num=100; $num>0; $num-=5){
                        if ($rand < $num){
                            $humi_stat{$num}{$sup}{"$bc-$umi"}++ ;
                            $hgene_stat{$num}{$sup}{$gene}++ ;
                            $hreads_stat{$num}{$sup}++ ;
                        }
                        else{
                            last ;
                        }
                    }
                }
                $process_reads += $reads ;
                while($process_reads/$sum_reads > $perc){
                    print "process: ", $perc*100, "%\n" ;
                    $perc += 0.05 ;
                }
            }
        }
    }

    open (OUT, ">$outfile") || die "$outfile, $!\n" ;
    print OUT "#Percent\tRawReadsNum\tSeqSaturation\tMeanReadsPerSupSpot\tMedianReads\tMedianUMI\tMedianGene\n" ;
    for my $percent (sort {$a<=>$b} keys %humi_stat){
        my @umi_nums = ();
        my @gene_nums = ();
        my @reads_nums = ();
        my $select_reads = 0 ;
        my $select_umi = 0 ;
        my $supspot_num = 0 ;
        for my $sup (keys %{$humi_stat{$percent}}){
            my $umi_num = scalar(keys %{$humi_stat{$percent}{$sup}}) ;
            my $gene_num = scalar(keys %{$hgene_stat{$percent}{$sup}});
            my $reads_num = $hreads_stat{$percent}{$sup} ;
            $select_reads += $reads_num ;
            $select_umi += $umi_num ;
            $supspot_num ++ ;
            push @umi_nums, $umi_num ;
            push @gene_nums, $gene_num ;
            push @reads_nums, $reads_num ;
        }
        my $seq_saturation = 1 - $select_umi/$select_reads ;
        my $median_umi = (sort {$a<=>$b} @umi_nums)[int($#umi_nums/2)] ;
        my $median_gene = (sort {$a<=>$b} @gene_nums)[int($#gene_nums/2)] ;
        my $median_reads = (sort {$a<=>$b} @reads_nums)[int($#reads_nums/2)] ;
        my $ave_reads = int($select_reads/$supspot_num+0.5) ;
        my $raw_reads = int($select_reads/$sum_reads*$total_reads) ;
        print OUT "$percent\t$raw_reads\t$seq_saturation\t$ave_reads\t$median_reads\t$median_umi\t$median_gene\n" ;
    }
    close(OUT);

    return ;
}

