#!/usr/bin/perl -w
# 
# Copyright (c) BIO_MARK 2021
# Writer:         Dengdj <dengdj@biomarker.com.cn>
# Program Date:   2021
# Modifier:       Dengdj <dengdj@biomarker.com.cn>
# Last Modified:  2021
my $ver="3.0";

use strict;
use Getopt::Long;
use Data::Dumper;
use FindBin qw($Bin $Script);
use File::Basename qw(basename dirname);

#####################

my %opts;
GetOptions(\%opts,"i=s","g=s","o=s","e","h" );

#&help()if(defined $opts{h});
if(!defined($opts{i}) || !defined($opts{g}) || !defined($opts{o}) || defined($opts{h}))
{
	print <<"	Usage End.";
	Description:
		
		Version: $ver
		   v1.2: change infile to bam.
		         change to use cds instead of gene.
		   v1.3: add gene map output.
		   v1.4: add unmap gene reads output.
		   v1.5: do not consider cds map.
		         consider ncRNA_gene etc.(col 3 ~/gene/)
		   v1.6: memory optimization.
		   v2.0: add map number info.  
		   v2.5: memory optimization.
		         one reads can appear at mult-line.
		   v2.6: support gtf file.   
		   v3.0: select MapQ 255 reads for analysis.
		         change the way for reads select.

	Usage:

		-i           reads mapping bam file                <infile>     must be given
		-g           gene gff file                         <infile>     must be given
		-o           result file                           <outfile>    must be given
		-e           only use exon(exclude intron)         [null]       optional     
		-h           Help document

	Usage End.

	exit;
}

###############Time
my $Time_Start;
$Time_Start = sub_format_datetime(localtime(time()));
&show_log("Start Time :[$Time_Start]");
################
$| = 1 ;
## get parameters
my $bamfile = $opts{i} ;
my $gfffile = $opts{g} ;
my $outfile = $opts{o} ;

## get gene info
my %hgene = ();
my %hcds = ();
if ($gfffile =~ /\.gff$|\.gff3$/){
    &get_gene_info($gfffile, \%hgene, \%hcds);
}
elsif($gfffile =~ /\.gtf$/){
    &get_gtf_gene_info($gfffile, \%hgene, \%hcds);
}

## get reads mapping info
my $amap_ids = &get_reads_mapping_info($bamfile, \%hgene, \%hcds, $outfile);

## out put stat file
&stat_and_output($amap_ids, $outfile);

###############Time
my $Time_End;
$Time_End = sub_format_datetime(localtime(time()));
&show_log("End Time :[$Time_End]");

###############Subs
sub sub_format_datetime {#Time calculation subroutine
    my($sec, $min, $hour, $day, $mon, $year, $wday, $yday, $isdst) = @_;
	$wday = $yday = $isdst = 0;
    sprintf("%4d-%02d-%02d %02d:%02d:%02d", $year+1900, $mon+1, $day, $hour, $min, $sec);
}

sub ABSOLUTE_DIR
{ #$pavfile=&ABSOLUTE_DIR($pavfile);
	my $cur_dir=`pwd`;
	$cur_dir =~ s/\n$//;
	my ($in)=@_;
	my $return="";
	
	if(-f $in)
	{
		my $dir=dirname($in);
		my $file=basename($in);
		chdir $dir;$dir=`pwd`;
		$dir =~ s/\n$// ;
		$return="$dir/$file";
	}
	elsif(-d $in)
	{
		chdir $in;$return=`pwd`;
		$return =~ s/\n$// ;
	}
	else
	{
		warn "Warning just for file and dir [$in]\n";
		exit;
	}
	
	chdir $cur_dir;
	return $return;
}

# &show_log("txt")
sub show_log()
{
	my ($txt) = @_ ;
	my $time = time();
	my $Time = &sub_format_datetime(localtime($time));
	print "$Time:\t$txt\n" ;
	return ($time) ;
}

#&run_or_die($cmd);
sub run_or_die()
{
	my ($cmd) = @_ ;
	my $start_time = &show_log($cmd);
	my $flag = system($cmd) ;
	if ($flag != 0){
		my $end_time = &show_log("Error: command fail: $cmd");
		&show_log("Elaseped time: ".($end_time - $start_time)."s\n");
		exit(1);
	}
	my $end_time = &show_log("done.");
	&show_log("Elaseped time: ".($end_time - $start_time)."s\n");
	return ;
}

## qsub
sub qsub()
{
	my ($shfile, $maxproc) = @_ ;
	my $dir = dirname($shfile) ;
	my $cmd = "cd $dir && sh /share/nas2/genome/bmksoft/tool/qsub_sge_plus/v1.0/qsub_sge.plus.sh --maxproc $maxproc --reqsub --independent $shfile" ;
	&run_or_die($cmd);

	return ;
}

## qsub_mult($shfile, $max_proc, $job_num)
sub qsub_mult()
{
	my ($shfile, $max_proc, $job_num) = @_ ;
	if ($job_num > 500){
		my @shfiles = &cut_shfile($shfile);
		for my $file (@shfiles){
			&qsub($file, $max_proc);
		}
	}
	else{
		&qsub($shfile, $max_proc) ;
	}
}

#my @shfiles = &cut_shfile($shfile);
sub cut_shfile()
{
	my ($file) = @_ ;
	my @files = ();
	my $num = 0 ;
	open (IN, $file) || die "Can't open $file, $!\n" ;
	(my $outprefix = $file) =~ s/.sh$// ;
	while (<IN>){
		chomp ;
		if ($num % 500 == 0){
			close(OUT);
			my $outfile = "$outprefix.sub_".(int($num/500)+1).".sh" ;
			push @files, $outfile ;
			open (OUT, ">$outfile") || die "Can't creat $outfile, $!\n" ;
		}
		print OUT $_, "\n" ;
		$num ++ ;
	}
	close(IN);
	close(OUT);

	return @files ;
}

sub sub_normal_dis_ran(){#$average,$standard_deviation
        my ($aver,$stand_dev) = @_ ;
        my $ran1 = rand() ;
        my $ran2 = rand() ;
        my $y = ((-2*log($ran1))**(0.5))*sin(2*3.1415926535*$ran2) ;
        return ($aver + $stand_dev*$y) ;
}

#&get_gtf_gene_info($gfffile, \%hgene, \%hcds);
sub get_gtf_gene_info()
{
    my ($gfffile, $ahgene, $ahcds) = @_ ;

    open (IN, $gfffile) || die "$gfffile, $!\n" ;
    my %hexon2gene = ();
    while(<IN>){
        chomp ;
        next if (m/^\#|^\s*$/);
        my ($chr, $soft, $type, $start, $end, undef, $ori, $offset, $info) = split(/\t+/, $_, 9) ;
        my @infos = split /\s+/, $info ;
        my %hinfos = ();
        for (my $i=0; $i<@infos; $i+=2){
            $hinfos{$infos[$i]} = $infos[$i+1] ;
        }
        my $gene_id = $hinfos{"gene_id"} ;
        $gene_id =~ s/[\"\;]//g ;
        if ($type =~ m/gene/i){
            next if ($type =~ /segment/) ;
            $ahgene->{$gene_id} = [$start, $end, $ori] ;
        }
        elsif ($type eq 'exon'){
            my $s_index = int($start/10000) ;
            my $e_index = int($end/10000) ;
            if ($s_index == $e_index){
                push @{$ahcds->{$chr}{$s_index}{$gene_id}}, [$start, $end, $ori] ;
            }
            else{
                if ($e_index - $s_index > 1){
                    for (my $i=$s_index; $i<=$e_index; $i++){
                        push @{$ahcds->{$chr}{$i}{$gene_id}}, [$start, $end, $ori] ;
                    }
                }
                else{
                    push @{$ahcds->{$chr}{$s_index}{$gene_id}}, [$start, $end, $ori] ;
                    push @{$ahcds->{$chr}{$e_index}{$gene_id}}, [$start, $end, $ori] ;
                }
            }
            if (!defined $hexon2gene{$gene_id}){
                $hexon2gene{$gene_id} = [$start, $end, $ori] ;
            }
            else{
                if ($hexon2gene{$gene_id}[0] > $start){
                    $hexon2gene{$gene_id}[0] = $start ;
                }
                if ($hexon2gene{$gene_id}[1] < $end){
                    $hexon2gene{$gene_id}[1] = $end ;
                }
            }
        }
    }
    close(IN);

    &cds_merge2($ahcds);
    if (scalar keys %{$ahgene} == 0){
        $ahgene = \%hexon2gene ;
    }

    return ;
}

#&get_gene_info($gfffile, \%hgene, \%hcds);
sub get_gene_info()
{
    my ($gfffile, $ahgene, $ahcds) = @_ ;

    open (IN, $gfffile) || die "$gfffile, $!\n" ;
    my $gene_id ;
    while(<IN>){
        chomp ;
        next if (m/^\#|^\s*$/);
        my ($chr, $soft, $type, $start, $end, undef, $ori, $offset, $info) = split ;
        if ($type =~ m/gene/i){
            next if ($type =~ /segment/) ;
            if ($info =~ m/ID=(.*?)\;/){
                $gene_id = $1 ;
            }
            elsif ($info =~ m/ID=([^;]*?)\s*$/){
                $gene_id = $1 ;
            }
            $gene_id =~ s/^gene[\:\-\_]// ;
            $ahgene->{$gene_id} = [$start, $end, $ori] ;
        }
        elsif ($type eq 'exon'){
            my $s_index = int($start/10000) ;
            my $e_index = int($end/10000) ;
            if ($s_index == $e_index){
                push @{$ahcds->{$chr}{$s_index}{$gene_id}}, [$start, $end, $ori] ;
            }
            else{
                if ($e_index - $s_index > 1){
                    for (my $i=$s_index; $i<=$e_index; $i++){
                        push @{$ahcds->{$chr}{$i}{$gene_id}}, [$start, $end, $ori] ;
                    }
                }
                else{
                    push @{$ahcds->{$chr}{$s_index}{$gene_id}}, [$start, $end, $ori] ;
                    push @{$ahcds->{$chr}{$e_index}{$gene_id}}, [$start, $end, $ori] ;
                }
            }
        }
    }
    close(IN);

    &cds_merge2($ahcds);

    return ;
}

#&cds_merge2($ahcds);
sub cds_merge2()
{
    my ($ahcds) = @_ ;

    for my $chr (keys %{$ahcds}){
        for my $index (keys %{$ahcds->{$chr}}){
            for my $id (keys %{$ahcds->{$chr}{$index}}){
                my @exon_reg = sort {$a->[0]<=>$b->[0] || $a->[1]<=>$b->[1]} @{$ahcds->{$chr}{$index}{$id}} ;
                my @regions = () ;
                push @regions, [@{$exon_reg[0]}] ;
                for (my $i=1; $i<@exon_reg; $i++){
                    my ($start, $end) = @{$exon_reg[$i]} ;
                    if ($start < $regions[-1][1]){
                        $regions[-1][1] = $end > $regions[-1][1] ? $end : $regions[-1][1] ;
                    }
                    else{
                        push @regions, [$start, $end] ;
                    }
                }
                $ahcds->{$chr}{$index}{$id} = \@regions ;
            }
        }
    }

    return ;
}

#my $amap_ids = &get_reads_mapping_info($bamfile, \%hgene, \%hcds, $outfile);
sub get_reads_mapping_info()
{
    my ($bamfile, $ahgene, $ahcds, $outfile) = @_ ;

    open (IN, "samtools view $bamfile|") || die "$bamfile, $!\n" ;
    open (EXN, ">$outfile.map2exon") || die "$outfile.map2exon, $!\n" ;
    open (ITR, ">$outfile.map2intron") || die "$outfile.map2intron, $!\n" ;
    open (TRA, ">$outfile.map2transcript") || die "$outfile.map2transcript, $!\n" ;
    open (ANT, ">$outfile.antisense") || die "$outfile.antisense, $!\n" ;
    my @maps_id = ();
    while(<IN>){
        chomp ;
        #next if (m/^\@/);
        my ($id, $flag, $chr, $start, $mapQ, $cigar, undef, undef, undef, $seq, $qual, $NH, $HI, $AS, $nM) = split ;
        $maps_id[$id] ||= "MP" ;
        next if ($mapQ != 255);
        $maps_id[$id] = "CM" ;
        next if (!defined $ahcds->{$chr});
        my $map_ori = "+" ;
        #print $flag & 16, "\n" ;
        if (($flag & 16) == 16){
            $map_ori = "-" ;
        }
        #print "$id\t$flag\t$map_ori\n" ;
        my $seqlen = length($seq) ;
        my ($maplen, $read_maplen, @matchs) = &get_maplen($cigar);
        my %hexon = ();
        my %hintron = ();
        my $gene_flag = 0 ;
        for (my $i=0; $i<@matchs; $i++){
            $matchs[$i][0] += $start - 1 ;
            $matchs[$i][1] += $start - 1 ;
            my $s_index = int($matchs[$i][0]/10000) ;
            my $e_index = int($matchs[$i][1]/10000) ;
            if ($s_index == $e_index){
                my $map_flag = &get_map_type($matchs[$i][0], $matchs[$i][1], $ahcds->{$chr}{$s_index}, $ahgene, \%hexon, \%hintron);
                if ($map_flag == 1){
                    $gene_flag = 1 ;
                }
            }
        }
        next if ($gene_flag == 0) ;
        my $antisense_flag = 0 ;
        my $map_flag = 0 ;
        if (scalar keys %hexon > 0){
            ($map_flag, $antisense_flag, my $aselect_genes) = &judge_exon_gene(\%hexon, $seqlen, $map_ori, $ahgene);
            if ($map_flag == 1){
                if (scalar @{$aselect_genes} > 1){
                    print EXN "$id\t@{$aselect_genes}\n" ;
                    $maps_id[$id] = 'EX-M' ;
                }
                else{
                    print EXN "$id\t",$aselect_genes->[0], "\n" ;
                    print TRA "$id\tEXON\t",$aselect_genes->[0], "\n" ;
                    $maps_id[$id] = 'EX-T' ;
                }
            }
        }
        if ($map_flag == 0){
            my ($intron_flag, $ant_flag, $aselect_genes, $acandi_genes) = &judge_intron_gene(\%hintron, $seqlen, $map_ori, $ahgene);
            if ($intron_flag == 1){
                $map_flag = 1 ;
                $maps_id[$id] = 'IN-U' ;
                if (scalar @{$aselect_genes} > 1){
                    print ITR "$id\t@{$aselect_genes}\n" ;
                }
                elsif (scalar @{$aselect_genes} == 1){
                    print ITR "$id\t",$aselect_genes->[0],"\n" ;
                    if (!defined $opts{e}){
                        print TRA "$id\tINTRON\t", $aselect_genes->[0],"\n" ;
                        $maps_id[$id] = 'IN-T' ;
                    }
                }
                elsif (scalar @{$acandi_genes} == 1){
                    print ITR "$id\t",$acandi_genes->[0],"\n" ;
                    if (!defined $opts{e}){
                        print TRA "$id\tINTRON\t", $acandi_genes->[0],"\n" ;
                        $maps_id[$id] = 'IN-T' ;
                    }
                }
                else{
                    print ITR "$id\t@{$acandi_genes}\n" ;
                }
            }
            elsif ($ant_flag){
                if ($antisense_flag == 1){
                    $maps_id[$id] = 'A-EX' ;
                    print ANT "$id\tEXON\n" ;
                }
                else{
                    $maps_id[$id] = 'A-IN' ;
                    print ANT "$id\tINTRON\n" ;
                }
            }
        }
    }
    close(IN);
    close(EXN);
    close(ITR);
    close(TRA);
    close(ANT);

    return(\@maps_id);
}

#my ($intron_flag, $ant_flag, $aselect_genes, $acandi_genes) = &judge_intron_gene(\%hintron, $seqlen, $map_ori, $ahgene);
sub judge_intron_gene()
{
    my ($ahintron, $seqlen, $map_ori, $ahgene) = @_ ;

    my $flag = 0 ;
    my @gene_ids = sort {$ahintron->{$b}<=>$ahintron->{$a} || $a cmp $b} keys %{$ahintron} ;
    my @select_genes = ();
    my @candi_genes = ();
    my $antisense_flag = 0 ;
    for my $gene_id (@gene_ids){
        my $intron_map_len = $ahintron->{$gene_id} ;
        my $gene_ori = $ahgene->{$gene_id}[2] ;
        if ($gene_ori ne $map_ori){
            $antisense_flag = 1 ;
            next ;
        }
        $flag = 1 ;
        if ($intron_map_len/$seqlen >= 0.5){
            push @select_genes, $gene_id ;
        }
        else{
            push @candi_genes, $gene_id ;
        }
    }

    return($flag, $antisense_flag, \@select_genes, \@candi_genes);
}

#($map_flag, $antisense_flag, my $aselect_genes) = &judge_exon_gene(\%hexon, $seqlen, $map_ori, $ahgene);
sub judge_exon_gene()
{
    my ($ahexon, $seqlen, $map_ori, $ahgene) = @_ ;

    my @gene_ids = sort {$ahexon->{$b}<=>$ahexon->{$a} || $a cmp $b} keys %{$ahexon} ;
    my $flag = 0 ;
    my $antisense_flag = 0 ;
    my @select_genes = () ;
    for my $gene_id (@gene_ids){
        my $exon_map_len = $ahexon->{$gene_id} ;
        my $gene_ori = $ahgene->{$gene_id}[2] ;
        if ($gene_ori ne $map_ori){
            $antisense_flag = 1 ;
            next ;
        }
        if ($exon_map_len/$seqlen >= 0.5){
            push @select_genes, $gene_id ;
            $flag = 1 ;
        }
        else{
            last ;
        }
    }

    return($flag, $antisense_flag, \@select_genes);
}

#&get_map_type($matchs[$i][0], $matchs[$i][1], $ahcds->{$chr}{$s_index}, $ahgene, \%hexon, \%hintron);
sub get_map_type()
{
    my ($start, $end, $ahcds_region, $ahgene, $ahexon, $ahintron) = @_ ;

    my $flag = 0 ;
    for my $gene_id (keys %{$ahcds_region}){
        my $map_len = 0 ;
        for (my $i=0; $i<@{$ahcds_region->{$gene_id}}; $i++){
            my ($exon_start, $exon_end, $ori) = @{$ahcds_region->{$gene_id}[$i]} ;
            if ($end < $exon_start){
                last ;
            }
            elsif ($start < $exon_end){
                if ($exon_end > $end){
                    $map_len += ($start < $exon_start) ? $end - $exon_start + 1 : $end - $start + 1 ;
                }
                else{
                    $map_len += ($start < $exon_start) ? $exon_end - $exon_start + 1 : $exon_end - $start + 1 ;
                }
            }
            else{
                next ;
            }
        }
        if ($map_len > 0){
            $ahexon->{$gene_id} += $map_len ;
            $flag = 1 ;
        }
        my ($gene_start, $gene_end, $gene_ori) = @{$ahgene->{$gene_id}} ;
        if ($end > $gene_start && $gene_end > $start){
            if ($gene_end > $end){
                $ahintron->{$gene_id} += ($start < $gene_start) ? $end - $gene_start + 1 : $end - $start + 1 ;
            }
            else{
                $ahintron->{$gene_id} += ($start < $gene_start) ? $gene_end - $gene_start + 1 : $gene_end - $start + 1 ;
            }
            $flag = 1 ;
        }
    }

    return($flag);
}

#my ($maplen, @matchs) = &get_maplen($cigar);
sub get_maplen()
{
    my ($cigar) = @_ ;

    my $len = 0 ;
    my @matchs = ();
    my $read_maplen = 0 ;
    while($cigar =~ m/(\d+)([a-zA-Z])/g){
        my ($num, $type) = ($1, $2) ;
        if ($type eq 'M' || $type eq '='){
            my $start = $len + 1 ;
            $len += $num ;
            my $end = $len ;
            push @matchs, [$start, $end] ;
            $read_maplen += $num ;
        }
        elsif ($type eq 'N' || $type eq 'D' || $type eq 'X'){
            $len += $num ;
        }
    }

    return($len, $read_maplen, @matchs);
}

#my $flag = &judge_cds($ahcds->{$g_id}, \@r_matchs);
sub judge_cds()
{
    my ($acds, $ar_matchs) = @_ ;

    my $index = 0 ;
    my $flag = 0 ;
    my $match_len = 0 ;
    if (!defined $acds){
        return($flag, $match_len);
    }
    #print Dumper @{$acds} ;
    for (my $i=0; $i<@{$acds}; $i++){
        my ($c_start, $c_end) = @{$acds->[$i]} ;
        for (my $j=$index; $j<@{$ar_matchs}; $j++){
            my ($r_start, $r_end) = @{$ar_matchs->[$j]} ;
            if ($r_end <= $c_start){
                $index = $j ;
                next ;
            }
            elsif ($r_start < $c_end){
                my @arr = sort {$a<=>$b} ($r_start, $c_start, $r_end, $c_end);
                $match_len += $arr[2] - $arr[1] + 1 ;
                #print "@arr\t$match_len\n" ;
                $flag = 1 ;
            }
            else{
                last ;
            }
        }
    }

    return($flag, $match_len);
}

#$flag = &judge_gene($g_start, $g_end, \@r_matchs);
sub judge_gene()
{
    my ($g_start, $g_end, $ar_matchs) = @_ ;

    my $match_len = 0 ;
    my $flag = 0 ;
    for (my $i=0; $i<@{$ar_matchs}; $i++){
        my ($r_start, $r_end) = @{$ar_matchs->[$i]} ;
        if ($r_end < $g_start){
            next ;
        }
        elsif ($r_start <= $g_end){
            my @arr = sort {$a<=>$b} ($r_start, $g_start, $r_end, $g_end);
            $match_len += $arr[2] - $arr[1] + 1 ;
            $flag = 1 ;
        }
        else{
            last ;
        }
    }

    return($flag, $match_len);
}

#&stat_and_output($amap_ids, $outfile);
sub stat_and_output()
{
    my ($amap_ids, $outfile) = @_ ;

    my %hstat = ();
    for (my $i=0; $i<@{$amap_ids}; $i++){
        next if (!defined $amap_ids->[$i]) ;
        $hstat{$amap_ids->[$i]}++ ;
    }

    open (OUT, ">$outfile.stat") || die "$outfile.stat, $!\n" ;
    $hstat{'EX-T'} ||= 0 ;
    $hstat{'EX-M'} ||= 0 ;
    $hstat{'IN-T'} ||= 0 ;
    $hstat{'IN-U'} ||= 0 ;
    $hstat{'A-EX'} ||= 0 ;
    $hstat{'A-IN'} ||= 0 ;
    $hstat{'MP'} ||= 0 ;
    $hstat{'CM'} ||= 0 ;
    my $exon = $hstat{'EX-M'} + $hstat{'EX-T'} + $hstat{'A-EX'} ;
    my $intron = $hstat{'IN-T'} + $hstat{'IN-U'} + $hstat{'A-IN'} ;
    my $inter = $hstat{'CM'} ;
    my $transcript = $hstat{'EX-T'} + $hstat{'IN-T'} ;
    my $antisense = $hstat{'A-EX'} ;
    if (!defined $opts{e}){
        $antisense += $hstat{'A-IN'} ;
    }
    my $confmap = $exon + $intron + $inter ;
    my $map = $confmap + $hstat{'MP'} ;
    print OUT "Reads Mapped to Genome\t$map\n" ;
    print OUT "Reads Mapped Confidently to Genome\t$confmap\n" ;
    print OUT "Reads Mapped Confidently to Intergenic Regions\t$inter\n" ;
    print OUT "Reads Mapped Confidently to Intronic Regions\t$intron\n" ;
    print OUT "Reads Mapped Confidently to Exonic Regions\t$exon\n" ;
    print OUT "Reads Mapped Confidently to Transcriptome\t$transcript\n" ;
    print OUT "Reads Mapped Antisense to Gene\t$antisense\n" ;
    close(OUT);

    return ;
}

